//
//      Copyright (C) DataStax Inc.
//
//   Licensed under the Apache License, Version 2.0 (the "License");
//   you may not use this file except in compliance with the License.
//   You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS,
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and
//   limitations under the License.
//

using System;
using System.Collections.Generic;
using System.Net;
using System.Threading.Tasks;

using Cassandra.Connections;
using Cassandra.ExecutionProfiles;
using Cassandra.Metrics.Internal;
using Cassandra.Observers.Abstractions;

namespace Cassandra.SessionManagement
{
    /// <inheritdoc />
    /// <remarks>This is an internal interface designed to declare the internal methods that are called
    /// across multiple locations of the driver's source code.</remarks>
    internal interface IInternalSession : ISession
    {
        /// <summary>
        /// Unique Session ID generated by the driver.
        /// </summary>
        Guid InternalSessionId { get; }

        /// <summary>
        /// Initialize the session
        /// </summary>
        Task Init();

        /// <summary>
        /// Gets or creates the connection pool for a given host
        /// </summary>
        IHostConnectionPool GetOrCreateConnectionPool(Host host, HostDistance distance);

        /// <summary>
        /// Gets a snapshot of the connection pools
        /// </summary>
        IEnumerable<KeyValuePair<IPEndPoint, IHostConnectionPool>> GetPools();

        /// <summary>
        /// Gets the existing connection pool for this host and session or null when it does not exists
        /// </summary>
        IHostConnectionPool GetExistingPool(IPEndPoint address);

        void CheckHealth(Host host, IConnection connection);

        bool HasConnections(Host host);

        void OnAllConnectionClosed(Host host, IHostConnectionPool pool);

        /// <summary>
        /// Gets or sets the keyspace
        /// </summary>
        new string Keyspace { get; set; }

        IInternalCluster InternalCluster { get; }
        
        /// <summary>
        /// Fetches the request options that were mapped from the provided execution profile's name.
        /// </summary>
        IRequestOptions GetRequestOptions(string executionProfileName);

        /// <summary>
        /// Returns the number of connected nodes.
        /// </summary>
        int ConnectedNodes { get; }

        IMetricsManager MetricsManager { get; }

        IObserverFactory ObserverFactory { get; }

        Task<RowSet> ExecuteAsync(IStatement statement, IRequestOptions requestOptions);
    }
}